'''
Created on Feb 21th, 2011

@author: Dmytro Korsakov
'''
import os
import sys
import time
import json
import xmlrpclib
import collections
import prettytable

from optparse import OptionParser, TitledHelpFormatter
from pkg_resources import parse_version, working_set

from config import Environment
from api import ScalrConnection, ScalrAPIError
from api.view import TableViewer

progname = 'scalr'
DEFAULT_API_URL = 'https://api.scalr.net'

farmrole_json_keys = [
	"chef.attributes",
	"chef.runlist",
	"openstack.networks",
	"aws.vpc_subnet_id",
	"aws.security_groups.list",
	"cloudstack.security_groups.list",
	"openstack.security_groups.list",
]

farmrole_fpath_keys = [
	"chef.ssh_private_key",
	]

class ScalrError(BaseException):
	pass

class Command(object):
	name = ''
	help = ''
	
	config = None
	parser = None
	options = None
		
	def __init__(self, config, *args):
		self.config = config
		self.parser = OptionParser(usage='%s %s %s' % (progname, self.name, self.help))
		self.inject_options(self.parser)
		self.options = self.parser.parse_args(list(args))[0]
		#Allowing user to use name instead of id if possible
		if hasattr(self.options, 'farm_id') and not self.options.farm_id and self.options.farm_name:
			self.options.farm_id = self.connection.get_farm_id(self.options.farm_name)
			
			if not self.options.farm_id:
				raise ScalrError("Farm '%s' not found" % self.options.farm_name)

		if hasattr(self.options, 'farm_name') and not self.options.farm_name and not self.options.farm_id:
			pass  # When farm_id is not reqired

		elif hasattr(self.options, 'farm_name') and not self.options.farm_name:
			self.options.farm_name = self.connection.get_farm_name(self.options.farm_id)
			
	def run(self):
		pass
	
	def call_api_method(self, callback, *args, **kwargs):
		try:
			return callback(*args, **kwargs)
		except ScalrAPIError, e:
			print e
			sys.exit(1)
	
	def pretty(self, callback, *args, **kwargs):
		try:
			return TableViewer(callback(*args, **kwargs))
		except ScalrAPIError, e:
			return e
			sys.exit(1)
			
	def require(self, *args):
		if not all(args):
			print self.parser.format_help(TitledHelpFormatter())
			sys.exit(1)
	
	@classmethod
	def inject_options(cls, parser):
		pass	
		
	@classmethod
	def usage(cls):
		parser = OptionParser(usage='%s %s %s' % (progname, cls.name, cls.help)) 
		cls.inject_options(parser)
		return parser.format_help(TitledHelpFormatter())

	@property		
	def connection(self):
		return ScalrConnection(self.config.environment,
					logger=self.config.logger)


class FarmRoleParametersList(Command):
	name = 'list-farm-role-parameters'
	help = '-r farm-role-id'

	def __init__(self, config, *args):
		super(FarmRoleParametersList, self).__init__(config, *args)
		self.require(self.options.farm_role_id)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-r", "--farm-role-id", dest="farm_role_id", help="FarmRoleID")

	def run(self):
		print self.pretty(self.connection.list_farm_role_parameters, self.options.farm_role_id)


class FarmRoleUpdateParameterValue(Command):
	name = 'update-farm-role-parameter'
	help = '-r farm-role-id -n param-name -v param-value'

	def __init__(self, config, *args):
		super(FarmRoleUpdateParameterValue, self).__init__(config, *args)
		self.require(self.options.farm_role_id, self.options.param_name, self.options.param_value)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-r", "--farm-role-id", dest="farm_role_id", help="FarmRoleID")
		parser.add_option("-n", "--param-name", dest="param_name", help="Parameter Name")
		parser.add_option("-v", "--value", dest="param_value", help="Parameter Value")

	def run(self):
		args = (self.options.farm_role_id, self.options.param_name, self.options.param_value)
		print self.pretty(self.connection.update_farm_role_parameter, *args)

		
class ApacheVhostsList(Command):
	name = 'list-apache-virtual-hosts'
	
	def run(self):
		print self.pretty(self.connection.list_apache_virtual_hosts)


class DmCreateSource(Command):
	name = 'dm-create-source'
	help = '-t [http|svn] -u url [-l login -p password]'

	def __init__(self, config, *args):
		super(DmCreateSource, self).__init__(config, *args)
		self.require(self.options.type, self.options.url)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-t", "--type", dest="type", help="Source type [svn|http]")
		parser.add_option("-u", "--url", dest="url", help="Source URL")
		parser.add_option("-l", "--login", dest="auth_login", default=None, help="Source login")
		parser.add_option("-p", "--password", dest="auth_password", default=None, help="Source password")
	
	def run(self):
		args = (self.options.type, self.options.url, self.options.auth_login, self.options.auth_password)
		print self.pretty(self.connection.dm_create_source, *args)


class DmCreateApplication(Command):
	name = 'dm-create-application'
	help = '-n name -s source-id [-b pre_deploy_script -a post_deploy_script]'

	def __init__(self, config, *args):
		super(DmCreateApplication, self).__init__(config, *args)
		self.require(self.options.name, self.options.source_id)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-n", "--name", dest="name", help="Application name")
		parser.add_option("-s", "--source-id", dest="source_id", help="Source ID")
		parser.add_option("-b", "--pre-deploy script", dest="pre_deploy_script", default=None, help="Script to execute BEFORE deploy")
		parser.add_option("-a", "--post-deploy script", dest="post_deploy_script", default=None, help="Script to execute AFTER deploy")
	
	def run(self):
		args = (self.options.name, self.options.source_id, self.options.pre_deploy_script, self.options.post_deploy_script)
		print self.pretty(self.connection.dm_create_application, *args)
		

class DmDeployApplication(Command):
	name = 'dm-deploy-application'
	help = '{-a app-id | -n app-name} -r farm-role-id -p remote-path [--non-interactive]'

	test_url = 'https://scalr-test-deploy.googlecode.com/svn/trunk' 
	test_app = 'ScalrDemoApp'
	test_type = 'svn'
	def __init__(self, config, *args):
		
		super(DmDeployApplication, self).__init__(config, *args)
		if self.options.app_name == self.test_app:
			self.require(self.options.app_id or self.options.app_name, self.options.farm_role_id)
		else:
			self.require(self.options.app_id or self.options.app_name, self.options.farm_role_id, self.options.remote_path)

	@classmethod
	def inject_options(cls, parser):
		app_name_help = "The name of application can be used INSTEAD of ID. Use name %s to test deployment process on your role." % cls.test_app
		parser.add_option("-n", "--app-name", dest="app_name", default=None, help=app_name_help)
		parser.add_option("-a", "--app-id", dest="app_id", help="Application ID")
		parser.add_option("-r", "--farm-role-id", dest="farm_role_id", help="FarmRole ID")
		parser.add_option("-p", "--remote-path", dest="remote_path", default='/var/www/', help="Remote path where to deploy the app. Default is /var/www/")
		parser.add_option("--non-interactive", dest="non_interactive", action="store_true", help="Use this option to avoid checking farm state")
	
	def run(self):
		self.options.app_id = self.options.app_id or self.connection.get_application_id(self.options.app_name)
		self.options.app_name = self.options.app_name or self.connection.get_application_name(self.options.app_id)
		
		if self.options.app_name == self.test_app:
			self.options.app_name = self.test_app
			
			#check sources
			sources = [source for source in self.connection.dm_list_sources() if source.url==self.test_url]
			if sources:
				sid = sources[0].id
			else:
				result = self.call_api_method(self.connection.dm_create_source,self.test_type, self.test_url, None, None)
				print TableViewer(result)
				sid = result[0].source_id
	
			#check apps
			apps = [app for app in self.connection.dm_list_applications() if app.source_id==sid]
			if apps:
				self.options.app_id = apps[0].id
			else:
				result = self.call_api_method(self.connection.dm_create_application,self.options.app_name, sid, None, None)
				print TableViewer(result)
				self.options.app_id = result[0].app_id

		#deploy
		args = (self.options.app_id, self.options.farm_role_id, self.options.remote_path)
		tasks = self.call_api_method(self.connection.dm_deploy_application, *args)
		print TableViewer(tasks)
							

		failed_tasks = []
		done_tasks = []
		#check task status in loop
		if not self.options.non_interactive and not isinstance(tasks,ScalrAPIError) and tasks:
			for task in tasks:
				
				print 'Deploying %s on server %s (DeploymentTaskID: %s)' % (self.options.app_name, task.server_id, task.task_id)
				
				for attempt in range(35): #600s
					try:
						ts_list = self.connection.dm_get_deployment_task_status(task.task_id)
						
						sys.stdout.write('.')
						sys.stdout.flush()
												
						if ts_list:
							status = ts_list[0].status
						
							if status == 'deployed':
								task.status = 'deployed' #changing old status
								done_tasks.append(task)
								print "Application '%s' has been successfully deployed on server %s." % (self.options.app_name, task.server_id)
								break
							
							elif status == 'failed':
								task.status = 'failed' #changing old status
								failed_tasks.append(task) #task
								#show log
								try:
									log = self.connection.dm_get_deployment_task_log(task.task_id)
									print TableViewer(log)

									if log.scalr_objects:
										task.errmsg = log.scalr_objects[0].message
								finally:
									print 'Deployment process %s has failed on server %s.' % (task.task_id, task.server_id)
								break
							
							elif status == 'pending':
								pass
							elif status == 'deploying':
								pass
							else:
								print status
							
						time.sleep(attempt)
						
					except (KeyboardInterrupt, SystemExit):
						#stops cheking current server and proceeds to the next
						break
				else:
					print "Maximum number of attempts was reached. On server %s application has not been deployed." % task.server_id
				print ''

			if done_tasks:
				print 'Completed tasks:'
				print TableViewer(done_tasks)

			if failed_tasks:
				print 'Failed tasks:'
				print TableViewer(failed_tasks)
				sys.exit(1)
			else:
				print 'All deployment tasks have finished successfully.'
				
		
class DmDeployApplicationAlias(DmDeployApplication):
	name = 'deploy'
	help = DmDeployApplication.help + '\nThis is an alias to dm-deploy-application command.'
				
	
class DmListDeploymentTasks(Command):
	name = 'dm-list-deployment-tasks'
	help = '-r farm-role-id -a app-id -s server-id'

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-a", "--app-id", dest="app_id", help="Application ID")
		parser.add_option("-r", "--farm-role-id", dest="farm_role_id", help="FarmRole ID")
		parser.add_option("-s", "--server-id", dest="server_id", default=None, help="Instance ID")
	
	def run(self):
		args = ( self.options.farm_role_id, self.options.app_id, self.options.server_id)
		print self.pretty(self.connection.dm_list_deployment_tasks, *args)
		
		
class DmGetDeploymentTaskStatus(Command):
	name = 'dm-get-deployment-task-status'
	help = '-t task-id'

	def __init__(self, config, *args):
		super(DmGetDeploymentTaskStatus, self).__init__(config, *args)
		self.require(self.options.task_id)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-t", "--task-id", dest="task_id", help="Deployment task ID")
	
	def run(self):
		print self.pretty(self.connection.dm_get_deployment_task_status, self.options.task_id)		
		
		
class DmSourcesList(Command):
	name = 'dm-list-sources'
	
	def run(self):
		print self.pretty(self.connection.dm_list_sources)
		

class DmApplicationsList(Command):
	name = 'dm-list-applications'
	
	def run(self):
		print self.pretty(self.connection.dm_list_applications)


class DmGetDeploymentTaskLog(Command):
	name = 'dm-get-deployment-task-log'
	help = '-t task-id [-s start-from -l limit]'

	def __init__(self, config, *args):
		super(DmGetDeploymentTaskLog, self).__init__(config, *args)
		self.require(self.options.task_id)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-t", "--task-id", dest="task_id", default=None, help="Task ID")
		parser.add_option("-s", "--start-from", dest="start", default=None, help="Start from specified event number (Can be used for paging)")
		parser.add_option("-l", "--record-limit", dest="limit", default=None, help="Limit number of returned events (Can be used for paging)")
	
	def run(self):
		args = (self.options.task_id, self.options.start, self.options.limit)
		print self.pretty(self.connection.dm_get_deployment_task_log, *args)


class DNSZonesList(Command):
	name = 'list-dns-zones'
	
	def run(self):
		print self.pretty(self.connection.list_dns_zones)


class FarmsList(Command):
	name = 'list-farms'

	def run(self):
		print self.pretty(self.connection.list_farms)
		

class ScriptsList(Command):
	name = 'list-scripts'
	
	def run(self):
		print self.pretty(self.connection.list_scripts)
		
		
class DNSZoneRecordsList(Command):
	name = 'list-dns-zone-records'
	help = '-n name'

	def __init__(self, config, *args):
		super(DNSZoneRecordsList, self).__init__(config, *args)
		self.require(self.options.name)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-n", "--zone-name", dest="name", default=None, help="Zone (Domain) name")
	
	def run(self):
		print self.pretty(self.connection.list_dns_zone_records, self.options.name)


class EventsList(Command):
	name = 'list-events'
	help = '{-f farm-id | -n name} [-s start-from -l limit]'

	def __init__(self, config, *args):
		super(EventsList, self).__init__(config, *args)
		self.require(self.options.farm_id)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-f", "--farm-id", dest="farm_id", default=None, help="Farm ID")
		parser.add_option("-n", "--farm-name", dest="farm_name", default=None, help="The name of farm could be used INSTEAD of ID")
		parser.add_option("-s", "--start-from", dest="start", default=None, help="Start from specified event number (Can be used for paging)")
		parser.add_option("-l", "--record-limit", dest="limit", default=None, help="Limit number of returned events (Can be used for paging)")
	
	def run(self):
		args = (self.options.farm_id, self.options.start, self.options.limit)
		print self.pretty(self.connection.list_events, *args)


class LogsList(Command):
	name = 'list-logs'
	help = '{-f farm-id | -n name} [-i server-id -s start-from -l limit]'

	def __init__(self, config, *args):
		super(LogsList, self).__init__(config, *args)
		self.require(self.options.farm_id or self.options.farm_name)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-f", "--farm-id", dest="farm_id", default=None, help="Farm ID")
		parser.add_option("-n", "--farm-name", dest="farm_name", default=None, help="The name of farm could be used INSTEAD of ID")
		parser.add_option("-i", "--server-id", dest="server_id", default=None, help="Instance ID")
		parser.add_option("-s", "--start-from", dest="start", default=None, help="Start from specified event number (Can be used for paging)")
		parser.add_option("-l", "--record-limit", dest="limit", default=None, help="Limit number of returned events (Can be used for paging)")

	def run(self):
		args = (self.options.farm_id, self.options.server_id, self.options.start, self.options.limit)
		print self.pretty(self.connection.list_logs, *args)


class ScriptingLogsList(Command):
	name = 'list-scripting-logs'
	help = '{-f farm-id | -n name} [-i server-id -s start-from -l limit]'

	def __init__(self, config, *args):
		super(ScriptingLogsList, self).__init__(config, *args)
		self.require(self.options.farm_id or self.options.farm_name)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-f", "--farm-id", dest="farm_id", default=None, help="Farm ID")
		parser.add_option("-n", "--farm-name", dest="farm_name", default=None, help="The name of farm could be used INSTEAD of ID")
		parser.add_option("-i", "--server-id", dest="server_id", default=None, help="Instance ID")
		parser.add_option("-s", "--start-from", dest="start", default=None, help="Start from specified event number (Can be used for paging)")
		parser.add_option("-l", "--record-limit", dest="limit", default=None, help="Limit number of returned events (Can be used for paging)")

	def run(self):
		args = (self.options.farm_id, self.options.server_id, self.options.start, self.options.limit)
		print self.pretty(self.connection.scripting_list_logs, *args)


class RolesList(Command):
	name = 'list-roles'
	help = '[-p platform -n name -x prefix -i image-id]'

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-p", "--platform", dest="platform", default=None, help="Cloud platform")
		parser.add_option("-n", "--name", dest="name", default=None, help="List roles with specified role name.")
		parser.add_option("-x", "--prefix", dest="prefix", default=None, help="List all roles begins with specified prefix.")
		parser.add_option("-i", "--image-id", dest="image_id", default=None, help="List roles with specified image id.")
	
	def run(self):
		args = (self.options.platform, self.options.name, self.options.prefix, self.options.image_id)
		print self.pretty(self.connection.list_roles, *args)


class ScriptGetDetails(Command):
	name = 'get-script-details'
	help = '-s script-id'

	def __init__(self, config, *args):
		super(ScriptGetDetails, self).__init__(config, *args)
		self.require(self.options.script_id)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-s", "--script-id", dest="script_id", default=None, help="Script ID")
	
	def run(self):
		print self.pretty(self.connection.get_script_details, self.options.script_id)
		

class BundleTaskGetStatus(Command):
	name = 'get-bundle-task-status'
	help = '-i bundle-task-id'
	
	def __init__(self, config, *args):
		super(BundleTaskGetStatus, self).__init__(config, *args)
		self.require(self.options.bt_id)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-i", "--bundle-task-id", dest="bt_id", default=None, help="ID of the bundle task")
	
	def run(self):
		print self.pretty(self.connection.get_bundle_task_status, self.options.bt_id)
		

class FarmGetDetails(Command):
	name = 'list-farm-roles'
	help = '{-f farm-id | -n name}'

	def __init__(self, config, *args):
		super(FarmGetDetails, self).__init__(config, *args)
		self.require(self.options.farm_id or self.options.farm_name)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-f", "--farm-id", dest="farm_id", default=None, help="Farm ID")
		parser.add_option("-n", "--farm-name", dest="farm_name", default=None, help="The name of farm could be used INSTEAD of ID")
	
	def run(self):
		print self.pretty(self.connection.get_farm_details, self.options.farm_id)


class FarmRoleProperties(Command):
	name = 'get-farm-role-properties'
	help = '{-f farm-id | -n name}'

	def __init__(self, config, *args):
		super(FarmRoleProperties, self).__init__(config, *args)
		self.require(self.options.farm_id)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-f", "--farm-id", dest="farm_id", default=None, help="Farm ID")
		parser.add_option("-n", "--farm-name", dest="farm_name", default=None, help="The name of farm could be used INSTEAD of ID")
	
	def run(self):
		print self.pretty(self.connection.get_farm_role_properties, self.options.farm_id)	
		
		
class ServerList(Command):
	name = 'list-servers'
	help = '{-f farm-id | -n name} [-r farm-role-id]'

	def __init__(self, config, *args):
		super(ServerList, self).__init__(config, *args)
		self.require(self.options.farm_id)

	def parse_variables(self, vars):
		'''
		v1,v2 -> list(v1,v2)
		'''
		return vars.split(',') if vars else []

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-f", "--farm-id", dest="farm_id", default=None, help="Farm ID")
		parser.add_option("-n", "--farm-name", dest="farm_name", default=None, help="The name of farm could be used INSTEAD of ID")
		parser.add_option("-r", "--farm-role-id", dest="farm_role_id", default=None, help="Farm Role ID")
		parser.add_option("-c", "--columns", dest="columns", default=None, help="Columns to show. Example: -c ExternalIP,Status,Agent")
	
	def run(self):
		args = (self.options.farm_id, self.options.farm_role_id, self.parse_variables(self.options.columns))
		print self.pretty(self.connection.list_servers, *args)
		
				
class FarmGetStats(Command):
	name = 'get-farm-stats'
	help = '{-f farm-id | -n name} [-d date]'

	def __init__(self, config, *args):
		super(FarmGetStats, self).__init__(config, *args)
		self.require(self.options.farm_id)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-f", "--farm-id", dest="farm_id", default=None, help="Farm ID")
		parser.add_option("-n", "--farm-name", dest="farm_name", default=None, help="The name of farm could be used INSTEAD of ID")
		parser.add_option("-d", "--date", dest="date", default=None, help="Date (mm-yyyy)")
	
	def run(self):
		print self.pretty(self.connection.get_farm_stats, self.options.farm_id,self.options.date)	
		

class StatisticsGetGraphURL(Command):
	name = 'get-statistics-graph-url'
	help = '-o object-type -i object-id -n watcher-name -g graph-type'

	def __init__(self, config, *args):
		super(StatisticsGetGraphURL, self).__init__(config, *args)
		self.require(self.options.obj_type, self.options.obj_id, self.options.name, self.options.graph_type)

	@classmethod
	def inject_options(cls, parser):		
		id_help = "In case if object type is instance ObjectID shoudl be server id, in case if object type is role ObjectID should be role id \
		and in case if object type is farm ObjectID should be farm id"
		parser.add_option("-o", "--object-type", dest="obj_type", default=None, help="Object type. Valid values are: role, server or farm")
		parser.add_option("-i", "--object-id", dest="obj_id", default=None, help=id_help)
		parser.add_option("-n", "--watcher-name", dest="name", default=None, help="Watcher name could be CPU, NET, MEM or LA")
		parser.add_option("-g", "--graph-type", dest="graph_type", default=None, help="Graph type could be daily, weekly, monthly or yearly")
	
	def run(self):
		args = (self.options.obj_type, self.options.obj_id, self.options.name, self.options.graph_type)
		print self.pretty(self.connection.get_statistics_graph_URL, *args)	


class ServerTerminate(Command):
	name = 'terminate-server'
	help = '-i server-id [-d decrease-min-instances-setting]'

	def __init__(self, config, *args):
		super(ServerTerminate, self).__init__(config, *args)
		self.require(self.options.server_id)

	@classmethod
	def inject_options(cls, parser):		
		parser.add_option("-i", "--server-id", dest="server_id", default=None, help='ServerID')
		parser.add_option("-d", "--decrease-min-instances", dest="decrease", default=None, help='Decrease MinInstances setting for role (Default: 0)')
	
	def run(self):
		print self.pretty(self.connection.terminate_server, self.options.server_id, self.options.decrease)			


class ServerImageCreate(Command):
	name = 'create-server-image'
	help = '-i server-id -n role-name'

	def __init__(self, config, *args):
		super(ServerImageCreate, self).__init__(config, *args)
		self.require(self.options.server_id, self.options.name)

	@classmethod
	def inject_options(cls, parser):		
		parser.add_option("-i", "--server-id", dest="server_id", default=None, help='ServerID')
		parser.add_option("-n", "--role-name", dest="name", default=None, help='Name for the new role')
		
	def run(self):
		print self.pretty(self.connection.create_server_image, self.options.server_id, self.options.name)	
	
	
class ServerLaunch(Command):
	name = 'launch-server'
	help = '-i farm-role-id [--increase-max-instances]'

	def __init__(self, config, *args):
		super(ServerLaunch, self).__init__(config, *args)
		self.require(self.options.farm_role_id)

	@classmethod
	def inject_options(cls, parser):		
		parser.add_option("-i", "--farm-role-id", dest="farm_role_id", default=None, help='FarmRoleID on which scalr should launch new instance')
		parser.add_option("--increase-max-instances", dest="increase_max_instances", action="store_true", help='Increase min instances for this role')
	
	def run(self):
		print self.pretty(self.connection.launch_server, self.options.farm_role_id, self.options.increase_max_instances)
	
		
class FarmLaunch(Command):
	name = 'launch-farm'
	help = '{-f farm-id | -n farm-name} [--non-interactive]'

	def __init__(self, config, *args):
		super(FarmLaunch, self).__init__(config, *args)
		self.require(self.options.farm_id)

	@classmethod
	def inject_options(cls, parser):		
		parser.add_option("-f", "--farm-id", dest="farm_id", default=None, help="The ID of farm that you want to launch")
		parser.add_option("-n", "--farm-name", dest="farm_name", default=None, help="The name of farm could be used INSTEAD of ID")
		parser.add_option("--non-interactive", dest="non_interactive", action="store_true", help="Use this option to avoid checking farm state")
		
	def run(self):
		result = self.pretty(self.connection.launch_farm, self.options.farm_id)	
		
		if self.options.non_interactive:
			print result
		
		elif not isinstance(result,ScalrAPIError):
			print "Starting farm '%s'.." % self.options.farm_name,
			for seconds in waits(60):
				try:
					status = self.connection.get_farm_status(self.options.farm_id)
					sys.stdout.write('.')
					sys.stdout.flush()
					if status == 'Running':
						print 'Started.' 
						break
					time.sleep(seconds)
				except (KeyboardInterrupt, SystemExit):
					break
			else: 
				print 'Cannot start farm. Please try again.'

class FarmTerminate(Command):
	name = 'terminate-farm'
	help = '{-f farm-id | -n farm-name} -e keep-ebs, -i keep-eip -d keep-dns-zone [--non-interactive]'

	def __init__(self, config, *args):
		super(FarmTerminate, self).__init__(config, *args)
		self.require(self.options.farm_id or self.options.farm_name)

	@classmethod
	def inject_options(cls, parser):		
		parser.add_option("-f", "--farm-id", dest="farm_id", default=None, help="The ID of farm that you want to terminate")
		parser.add_option("-n", "--farm-name", dest="farm_name", default=None, help="The name of farm could be used INSTEAD of ID")
		parser.add_option("-e", "--keep-ebs", dest="keep_ebs", default=None, help="Keep EBS volumes created for roles on this farm 0|1")
		parser.add_option("-i", "--keep-eip", dest="keep_eip", default=None, help='Keep Elastic IPs created for roles on this farm 0|1')
		parser.add_option("-d", "--keep-dns-zone", dest="keep_dns_zone", default=None, help="Keep DNS zone that assigned to this farm on nameservers 0|1")
		parser.add_option("--non-interactive", dest="non_interactive", action="store_true", help="Use this option to avoid checking farm state")
	
	def run(self):
		args = (self.options.farm_id, self.options.keep_ebs, self.options.keep_eip, self.options.keep_dns_zone)
		result = self.pretty(self.connection.terminate_farm, *args)	

		if self.options.non_interactive:
			print result
		
		elif not isinstance(result,ScalrAPIError):
			print "Stopping farm '%s'.." % self.options.farm_name,
			for seconds in waits(60):
				try:
					status = self.connection.get_farm_status(self.options.farm_id)
					sys.stdout.write('.')
					sys.stdout.flush()
					if status == 'Stopped':
						print 'Stopped.'
						break
					time.sleep(seconds)
				except (KeyboardInterrupt, SystemExit):
					break
			else: 
				print 'Cannot stop farm. Please try again.'


class ScriptExecute(Command):
	name = 'execute-script'
	help = '{-f farm-id | -n name} -e script-id -a async -t timeout [-i farm-role-id -s server-id -r revision -v variables]'
	#TODO: Test passing variables  

	def __init__(self, config, *args):
		super(ScriptExecute, self).__init__(config, *args)
		self.require(self.options.farm_id, self.options.script_id, self.options.mode, self.options.timeout)

	@classmethod
	def inject_options(cls, parser):		
		id_help = "Script will be executed on all instances with this farm role ID. You can get this ID by using GetFarmDetails API call"
		parser.add_option("-i", "--farm-role-id", dest="farm_role_id", default=None, help=id_help)
		parser.add_option("-s", "--server-id", dest="server_id", default=None, help="Script will be executed on this server")
		parser.add_option("-f", "--farm-id", dest="farm_id", default=None, help="Execute script on specified farm")
		parser.add_option("-n", "--farm-name", dest="farm_name", default=None, help="The name of farm could be used INSTEAD of ID")
		parser.add_option("-e", "--script-id", dest="script_id", default=None, help="Script ID")
		parser.add_option("-t", "--timeout", dest="timeout", default=None, help="Script execution timeout (seconds)")
		parser.add_option("-a", "--async", dest="mode", default=None, help="Excute script asynchronously (1) or synchronously (0)")
		parser.add_option("-r", "--revision", dest="revision", default=None, help="Execute specified revision of script")
		parser.add_option("-v", "--variables", dest="variables", default=None, help="Script variables.Example: key1=value1,key2=value2")


	def run(self):
		args = (self.options.farm_id, self.options.script_id, self.options.timeout \
				, self.options.mode, self.options.farm_role_id, self.options.server_id, parse_kv_options(self.options.variables), self.options.revision)
		print self.pretty(self.connection.execute_script, *args)	


class FireCustomEvent(Command):
	name = 'fire-custom-event'
	help = '-s server-id -n event-name [-v variables]'

	def __init__(self, config, *args):
		super(FireCustomEvent, self).__init__(config, *args)
		self.require(self.options.server_id, self.options.event_name)

	@classmethod
	def inject_options(cls, parser):
		id_help = "Fire custom Scalr Event on the node"
		parser.add_option("-s", "--server-id", dest="server_id", default=None, help="The event will be fired from this server")
		parser.add_option("-n", "--event-name", dest="event_name", default=None, help="Name of the event to fire")
		parser.add_option("-v", "--variables", dest="variables", default=None, help="Event variables.Example: key1=value1,key2=value2")

	def parse_variables(self, vars):
		'''
		k1=v1;k2=v2 -> dict(k1=v1,k2=v2)
		'''
		s = {}
		if vars:
			for pair in vars.split(','):
				k,v = pair.split('=')
				s[k] = v
		return s

	def run(self):
		args = (self.options.server_id, self.options.event_name, self.parse_variables(self.options.variables))
		print self.pretty(self.connection.fire_custom_event, *args)


class ServerReboot(Command):
	name = 'reboot-server'
	help = '-s server-id'

	def __init__(self, config, *args):
		super(ServerReboot, self).__init__(config, *args)
		self.require(self.options.server_id)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-s", "--server-id", dest="server_id", default=None, help="Server ID")
	
	def run(self):
		print self.pretty(self.connection.reboot_server, self.options.server_id)	


class DNSZoneCreate(Command):
	name = 'create-dns-zone'
	help = '-d domain-name [{-f farm-id | -n farm-name} -i farm-role-id]'

	def __init__(self, config, *args):
		super(DNSZoneCreate, self).__init__(config, *args)
		self.require(self.options.name)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-d", "--domain-name", dest="name", default=None, help="Domain (Application) name")
		parser.add_option("-f", "--farm-id", dest="farm_id", default=None, help="Farm ID")
		parser.add_option("-n", "--farm-name", dest="farm_name", default=None, help="The name of farm could be used INSTEAD of ID")
		parser.add_option("-i", "--farm-role-id", dest="farm_role_id", default=None, help="Farm Role ID")
	
	def run(self):
		print self.pretty(self.connection.create_dns_zone, self.options.name, self.options.farm_id, self.options.farm_role_id)		
	
	
class DNSZoneRecordAdd(Command):
	name = 'add-dns-zone-record'
	help = '-z zone-name -t type -l ttl -n record-name -v record-value [-p priority -w weight -o port]'

	def __init__(self, config, *args):
		super(DNSZoneRecordAdd, self).__init__(config, *args)
		self.require(self.options.zone_name, self.options.type, self.options.ttl, self.options.record_name, self.options.record_value)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-z", "--zone-name", dest="zone_name", default=None, help="Zone (Domain) name")
		parser.add_option("-t", "--type", dest="type", default=None, help="Record type (A, MX, CNAME, NS, TXT, SRV)")
		parser.add_option("-l", "--ttl", dest="ttl", default=None, help="Record TTL")
		parser.add_option("-n", "--record-name", dest="record_name", default=None, help="Record name")
		parser.add_option("-v", "--record-value", dest="record_value", default=None, help="Record value")
		
		parser.add_option("-p", "--priority", dest="priority", default=None, help="Priority (for MX records)")
		parser.add_option("-w", "--weight", dest="weight", default=None, help="Weight (for SRV records)")
		parser.add_option("-o", "--port", dest="port", default=None, help="Port (for SRV records)")		
	
	def run(self):
		args = (self.options.zone_name, self.options.type, self.options.ttl, self.options.record_name, self.options.record_value,
				self.options.priority, self.options.weight, self.options.port)
		print self.pretty(self.connection.add_dns_zone_record, *args)	


class DNSZoneRecordRemove(Command):
	name = 'remove-dns-zone-record'
	help = '-n name -i record-id'

	def __init__(self, config, *args):
		super(DNSZoneRecordRemove, self).__init__(config, *args)
		self.require(self.options.name, self.options.record_id)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-n", "--domain-name", dest="name", default=None, help="Domain (Application) name")
		parser.add_option("-i", "--record-id", dest="record_id", default=None, help="Record ID")
	
	def run(self):
		print self.pretty(self.connection.remove_dns_zone_record, self.options.name, self.options.record_id)			

	
class ApacheVhostCreate(Command):
	name = 'create-apache-vhost'
	help = '-d domain-name {-f farm-id | -n name} -i farm-role-id -r document-root -s enable-ssl [-k key-path -c cert-path]'
	
	def __init__(self, config, *args):
		super(ApacheVhostCreate, self).__init__(config, *args)
		self.require(self.options.domain, self.options.farm_id, self.options.farm_role_id, self.options.document_root, self.options.enable_ssl)
		pass			

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-d", "--domain-name", dest="domain", default=None, help="Domain (Application) name")
		parser.add_option("-f", "--farm-id", dest="farm_id", default=None, help="Farm ID")
		parser.add_option("-n", "--farm-name", dest="farm_name", default=None, help="The name of farm could be used INSTEAD of ID")
		parser.add_option("-i", "--farm-role-id", dest="farm_role_id", default=None, help="Farm Role ID")
		parser.add_option("-r", "--document-root-dir", dest="document_root", default=None, help="Document root for virtualhost")
		parser.add_option("-s", "--enable-ssl", dest="enable_ssl", default=None, help="EnableSSL (1 - yes, 0 - no)")
		parser.add_option("-k", "--pk-path", dest="pk_path", default=None, help="Path to file with Private key for SSL")
		parser.add_option("-c", "--cert-path", dest="cert_path", default=None, help="Path to file with Certificate for SSL")
	
	def run(self):
		pk = None 
		cert = None
		
		if self.options.enable_ssl and int(self.options.enable_ssl):
			self.require(self.options.pk_path, self.options.cert_path)
			
			if not os.path.exists(self.options.cert_path):
				print '%s : file with certificate not found.' % self.options.cert_path
				sys.exit(1)
			else: 
				cert = open(self.options.cert_path, 'r').read()
				
			if not os.path.exists(self.options.pk_path):
				print '%s : file with private key not found.' % self.options.pk_path
				sys.exit(1)
			else: 
				pk = open(self.options.pk_path, 'r').read()
				
		args = (self.options.domain, self.options.farm_id, self.options.farm_role_id, self.options.document_root, self.options.enable_ssl, pk, cert)
		print self.pretty(self.connection.create_apache_vhost, *args)	
		
	
class ConfigureEnv(Command):
	name = 'configure'
	help = '-i key_id -a key -u api_url -e env_id'
		
	def __init__(self, config, *args):
		super(ConfigureEnv, self).__init__(config, *args)
		self.require(self.options.key_id, self.options.key, self.options.api_url)
			
	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-i", "--key-id", dest="key_id", default=None, help="Scalr API key ID")
		parser.add_option("-a", "--access-key", dest="key", default=None, help="Scalr API access key")
		parser.add_option("-e", "--env-id", dest="env_id", default=None, help="Scalr Environment ID")
		parser.add_option("-u", "--api-url", dest="api_url", default=DEFAULT_API_URL, help="Scalr API URL (IF you use open source Scalr installation)")
		
	def run(self):		
		e = Environment(url=self.options.api_url,
				key_id=self.options.key_id,
				key=self.options.key,
				env_id = self.options.env_id,
				api_version = '2.3.0',
				auth_type = 'password')
		
		e.write(self.config.base_path)
		e = Environment.from_ini(self.config.base_path)
		print e


class ConfigureLDAPEnv(Command):
	name = 'configure-ldap'
	help = '-l login -u -api_url -e env_id'

	def __init__(self, config, *args):
		super(ConfigureLDAPEnv, self).__init__(config, *args)
		self.require(self.options.login, self.options.api_url)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-l", "--login", dest="login", default=None, help="Login for LDAP authentication")
		parser.add_option("-e", "--env-id", dest="env_id", default=None, help="Scalr Environment ID")
		parser.add_option("-u", "--api-url", dest="api_url", default=DEFAULT_API_URL, help="Scalr API URL (IF you use open source Scalr installation)")

	def run(self):
		e = Environment(
				url=self.options.api_url,
				ldap_login=self.options.login,
				env_id = self.options.env_id,
				api_version = '2.3.0',
				auth_type='ldap')

		e.write(self.config.base_path)
		e = Environment.from_ini(self.config.base_path)
		print e

class EnvironmentsList(Command):
	name = 'list-environments'

	def run(self):
		print self.pretty(self.connection.list_environments)


class ShowConfig(Command):
	name = 'show-config'

	def run(self):
		e = Environment.from_ini(self.config.base_path)
		print e


class CheckUpdates(Command):
	name = 'check-updates'

	def run(self):
		versions = dict([(d.key, d.version) for d in working_set])
		installed = versions['scalr']
		print 'Installed version: %s' % (installed)

		repo = 'http://pypi.python.org/pypi'
		pypi = xmlrpclib.ServerProxy(repo)
		available = pypi.package_releases('scalr')

		if available:
			latest = available[-1]
			print 'Latest version in repo %s: %s' % (repo, latest)
			if cmp(parse_version(installed), parse_version(latest)):
				print 'Update available.'
				print "You can update scalr-tools by executing 'sudo easy_install --upgrade scalr' in Linux/OSX console."
			else:
				print 'No updates available.'


class FarmCreate(Command):
	name = 'create-farm'
	help = '-n name [-d description]'

	def __init__(self, config, *args):
		super(FarmCreate, self).__init__(config, *args)
		self.require(self.options.name)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-n", "--farm-name", dest="name", default=None, help="A name for the Farm to be created")
		help_descr = "A description for the Farm to be created. Interactive mode required"
		parser.add_option("-d", "--description", dest="descr", action="store_true", help=help_descr)
		proj_id = "The ID of the Cost Analytics Project to assign this Farm to. Rquired, if Cost Analytics is in use"
		parser.add_option("-i", "--project-id", dest="project_id", default=None, help=proj_id)
		parser.add_option("-c", "--configuration", dest="config", default=None, help="Farm configuration. Example: vpc_region=us-east-1,vpc_id=vpc-8371b8e6")

	def run(self):
		descr = None
		if not self.options.descr:
			descr = raw_input("Farm description []:")
		print self.pretty(self.connection.create_farm, self.options.name, descr, self.options.project_id, parse_kv_options(self.options.config))


class FarmRemove(Command):
	name = 'remove-farm'
	help = '{-f farm-id | -n farm-name}'

	def __init__(self, config, *args):
		super(FarmRemove, self).__init__(config, *args)
		self.require(self.options.farm_id or self.options.farm_name)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-f", "--farm-id", dest="farm_id", default=None, help="ID of the farm that you want to terminate")
		parser.add_option("-n", "--farm-name", dest="farm_name", default=None, help="Farm name could be used INSTEAD of ID")

	def run(self):
		print self.pretty(self.connection.remove_farm, self.options.farm_id)


class FarmRemoveRole(Command):
	name = 'remove-farm-role'
	help = '{-f farm-id | -n name} -r farm-role-id'

	def __init__(self, config, *args):
		super(FarmRemoveRole, self).__init__(config, *args)
		self.require(self.options.farm_id, self.options.farm_role_id)

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-f", "--farm-id", dest="farm_id", default=None, help="Farm ID")
		parser.add_option("-n", "--farm-name", dest="farm_name", default=None, help="The name of farm could be used INSTEAD of ID")
		parser.add_option("-r", "--farm-role-id", dest="farm_role_id", default=None, help="Farm Role ID")

	def run(self):
		args = (self.options.farm_id, self.options.farm_role_id)
		print self.pretty(self.connection.remove_farm_role, *args)


class FarmAddRole(Command):
	name = 'add-role-to-farm'
	help = '{-f farm-id | -n name} -r role_id -p platform -l cloud-location -a alias [-c configuration]'

	def __init__(self, config, *args):
		super(FarmAddRole, self).__init__(config, *args)
		self.require(self.options.farm_id, self.options.role_id, self.options.platform, self.options.location, self.options.alias)

	@classmethod
	def inject_options(cls, parser):
		list_opts = ", ".join(sorted(farmrole_json_keys+farmrole_fpath_keys))
		configuration_help = "Configuration for the Farm Role. Example: key1=value1,key2=value2. "
		configuration_help += "Certain parameters should be passed through a file "
		configuration_help += "(e.g. -c cloudstack.security_groups.list=/path/to/file/cs_sgroups.json) "
		configuration_help += "Those parameters are: \n %s" % list_opts
		parser.add_option("-f", "--farm-id", dest="farm_id", default=None, help="The ID of the Farm you'd like to add a Role to")
		parser.add_option("-n", "--farm-name", dest="farm_name", default=None, help="The name of farm could be used INSTEAD of ID")
		parser.add_option("-r", "--role-id", dest="role_id", default=None, help="The ID of the Role you'd like to add to the Farm")
		parser.add_option("-p", "--platform", dest="platform", default="ec2", help="https://scalr-wiki.atlassian.net/wiki/display/docs/API+Constants")
		parser.add_option("-l", "--location", dest="location", default=None, help="The identifier of the Cloud Location the Farm Role should be launched in")
		parser.add_option("-a", "--alias", dest="alias", default=None, help="Farm Role Alias. Must be longer than 4 characters, and should match: [a-zA-Z0-9-_]")
		parser.add_option("-c", "--configuration", dest="configuration", default=None, help=configuration_help)

	def run(self):
		args = (self.options.farm_id, self.options.role_id, self.options.platform, self.options.location,
										self.options.alias, prepare_farmrole_settings(parse_kv_options(self.options.configuration)))
		print self.pretty(self.connection.add_farm_role, *args)


class FarmUpdateRole(Command):
	name = 'update-farm-role'
	help = '-r farm-role-id [-a alias -c configuration]'

	def __init__(self, config, *args):
		super(FarmUpdateRole, self).__init__(config, *args)
		self.require(self.options.farm_role_id)

	@classmethod
	def inject_options(cls, parser):
		list_opts = ", ".join(sorted(farmrole_json_keys+farmrole_fpath_keys))
		configuration_help = "Configuration for the Farm Role. Example: key1=value1,key2=value2. "
		configuration_help += "Certain parameters should be passed through a file "
		configuration_help += "(e.g. -c cloudstack.security_groups.list=/path/to/file/cs_sgroups.json) "
		configuration_help += "Those parameters are: \n %s" % list_opts
		parser.add_option("-r", "--farm-role-id", dest="farm_role_id", default=None, help="Farm Role ID")
		parser.add_option("-a", "--alias", dest="alias", default=None, help="Farm Role Alias. Must be longer than 4 characters, and should match: [a-zA-Z0-9-_]")
		parser.add_option("-c", "--configuration", dest="configuration", default=None, help=configuration_help)

	def run(self):
		args = (self.options.farm_role_id, self.options.alias, prepare_farmrole_settings(parse_kv_options(self.options.configuration)))
		print self.pretty(self.connection.update_farm_role, *args)


class GlobalVariablesList(Command):
	name = 'list-global-variables'

	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-f", "--farm-id", dest="farm_id", default=None, help="The ID of the Farm you'd like to add a Role to")
		parser.add_option("-n", "--farm-name", dest="farm_name", default=None, help="The name of farm could be used INSTEAD of ID")
		parser.add_option("-i", "--farm-role-id", dest="farm_role_id", default=None, help="Farm Role ID")
		parser.add_option("-r", "--role-id", dest="role_id", default=None, help="The ID of the Role you'd like to add to the Farm")
		parser.add_option("-s", "--server-id", dest="server_id", default=None, help="Instance ID")


	def run(self):
		args = (self.options.role_id, self.options.farm_id, self.options.farm_role_id, self.options.server_id)
		print self.pretty(self.connection.list_global_variables, *args)


class ServerGetExtendedInformation(Command):
	name = 'get-server-extended-information'

	def __init__(self, config, *args):
		super(ServerGetExtendedInformation, self).__init__(config, *args)
		self.require(self.options.server_id)


	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-s", "--server-id", dest="server_id", default=None, help="Instance ID")


	def run(self):
		kv = self.connection.get_extended_server_information(self.options.server_id)
		pt = prettytable.PrettyTable(field_names=["Name", "Value"])
		for k,v in kv.items():
			pt.add_row([k,v])
		print pt


class GlobalVariableSet(Command):
	name = 'set-global-variable'


	def __init__(self, config, *args):
		super(GlobalVariableSet, self).__init__(config, *args)
		self.require(self.options.param_name, self.options.param_value,
					 self.options.farm_id or self.options.farm_role_id or self.options.role_id or self.options.server_id)


	@classmethod
	def inject_options(cls, parser):
		parser.add_option("-k", "--param-name", dest="param_name", default=None, help="Global Variable Name.")
		parser.add_option("-v", "--param-value", dest="param_value", default=None, help="Global Variable Value.")
		parser.add_option("-f", "--farm-id", dest="farm_id", default=None, help="The ID of the Farm you'd like to add a Role to")
		parser.add_option("-n", "--farm-name", dest="farm_name", default=None, help="The name of farm could be used INSTEAD of ID")
		parser.add_option("-i", "--farm-role-id", dest="farm_role_id", default=None, help="Farm Role ID")
		parser.add_option("-r", "--role-id", dest="role_id", default=None, help="The ID of the Role you'd like to add to the Farm")
		parser.add_option("-s", "--server-id", dest="server_id", default=None, help="Instance ID")


	def run(self):
		args = [self.options.param_name, self.options.param_value]
		args += [self.options.role_id, self.options.farm_id, self.options.farm_role_id, self.options.server_id]
		print self.pretty(self.connection.set_global_variable, *args)

def waits(n):
	num = 0
	sum = 0
	while sum <= n:
		yield num
		num += 1
		sum += num


def parse_kv_options(vars):
	'''
	k1=v1;k2=v2 -> dict(k1=v1,k2=v2)
	'''
	s = {}
	if vars:
		for pair in vars.split(','):
			k,v = pair.split('=')
			s[k] = v
	return s


def prepare_farmrole_settings(data):
	d = {}
	for attribute, value in data.items():
		if attribute in farmrole_json_keys:
			value = os.path.realpath(os.path.expanduser(value))
			if not os.path.exists(value):
				raise ScalrError("Error: JSON configuration file %s not found." % value)
			try:
				with open(value) as fp:
					text = fp.read().strip()
					json.loads(text)
					d[attribute] = text
			except (TypeError, ValueError), e:
				raise ScalrError("Cannot parse JSON in %s: %s" % (value, str(e)))

		elif attribute in farmrole_fpath_keys:
			value = os.path.realpath(os.path.expanduser(value))
			if not os.path.exists(value):
				raise ScalrError("Error: %s not found." % value)
			try:
				with open(value) as fp:
					d[attribute] = convert(fp.read())
			except BaseException, e:
				raise ScalrError("Cannot read from file %s: %s" % (value, str(e)))

		else:
			d[attribute] = value
	return d


def convert(data):
	#converts unicode to string
	if isinstance(data, basestring):
		return str(data)
	elif isinstance(data, collections.Mapping):
		return dict(map(convert, data.iteritems()))
	elif isinstance(data, collections.Iterable):
		return type(data)(map(convert, data))
	else:
		return data


'''
Created on Feb 21th, 2011

@author: Dmytro Korsakov
'''

import binascii
import hashlib
import logging
import hmac
import sys


try:
	import timemodule as time
except ImportError:
	import time

import os
import getpass
from xml.dom.minidom import parseString
from urllib import urlencode, splitnport
from urllib2 import urlopen, Request, URLError, HTTPError

import types_ as types

class ScalrConnection(object):
	'''
	implements Scalr API
	'''

	def __init__(self, environment, logger=None):
		self.environment = environment
		'''
		self.url = url
		self.key_id  = key_id
		self.access_key = access_key
		self.env_id = env_id
		self.api_version = api_version
		'''
		self._logger = logger or logging.getLogger(__name__)
		self.last_transaction_id = None
		
		
	def fetch(self, command, **params):
		"""
		@return object
		"""
		# Perform HTTP request
		request_body = {}
		request_body["Action"] = command
		request_body["Version"] = self.environment.api_version
		request_body['AuthVersion'] = '3'
		if self.environment.env_id and self.environment.env_id != 'None':
			request_body['EnvID'] = self.environment.env_id
		#request_body['SysDebug'] = '1'

		if {} != params :
			for key, value in params.items():
				if isinstance(value, dict):
					for k,v in value.items():
						request_body['%s[%s]'%(key,k)] = v
				else:
					request_body[key] = value

		if not self.environment.auth_type or self.environment.auth_type == 'password':
			request_body["KeyID"] = self.environment.key_id
			signature, timestamp = sign_http_request_v3(request_body, self.environment.key_id, self.environment.key)
			request_body["TimeStamp"] = timestamp
			request_body["Signature"] = signature

		elif self.environment.auth_type == 'ldap':
			request_body["Login"] = self.environment.ldap_login
			request_body["AuthType"] = self.environment.auth_type

			password =self.environment.ldap_password or os.environ.get('SCALR_API_LDAP_PASSWORD')
			if not password:
				password = getpass.getpass('Password: ')
			request_body["Password"] = password

		post_data = urlencode(request_body)

		self._logger.debug('POST URL: \n%s' % self.environment.url)
		self._logger.debug('POST DATA: \n%s' % post_data)
		
		response = None
		try:
			req = Request(self.environment.url, post_data, {})
			response = urlopen(req)
		except URLError, e:
			if isinstance(e, HTTPError):
				resp_body = e.read() if e.fp is not None else ""
				raise ScalrAPIError("Request failed. %s. URL: %s. Service message: %s" % (e, self.environment.url, resp_body))
			else:
				host, port = splitnport(req.host, req.port or 443)
				raise ScalrAPIError("Cannot connect to Scalr on %s:%s. %s" 
						% (host, port, str(e)))
		except BaseException,e:
			raise ScalrAPIError("Cannot connect to Scalr: %s" % str(e))
			
		resp_body = response.read()
		self._logger.debug("SCALR RESPONSE: \n%s", resp_body)
		
		# Parse XML response
		xml = None
		try:
			xml = xml_strip(parseString(resp_body))
		except (Exception, BaseException), e:
			raise ScalrAPIError("Cannot parse XML. %s" % [str(e)])		
		
		self._logger.debug('VALID XML: \n%s' % xml.toprettyxml())
		
		return xml


	def list_farm_role_parameters(self, farm_role_id):
		"""
		@return FarmRoleParameter[]
		"""
		params = {'FarmRoleID':farm_role_id}

		return self._request(command="FarmRoleParametersList", params=params, response_reader=self._read_list_farm_role_parameters)


	def update_farm_role_parameter(self, farm_role_id, param_name, param_value):
		"""
		@return Result[]
		"""
		params = {}
		#, ,
		params['FarmRoleID'] = farm_role_id
		params['ParamName']  = param_name
		params['ParamValue'] 	= param_value

		return self._request(command="FarmRoleUpdateParameterValue", params=params, response_reader=self._read_update_farm_role_parameter)


	def dm_create_source(self, type, url, auth_login=None, auth_password=None):
		"""
		@return Result[]
		"""
		params = {}
		
		params['Type'] = type
		params['URL']  = url
		if auth_login	: params['AuthLogin'] 	= auth_login
		if auth_password: params['AuthPassword']= auth_password
		
		return self._request(command="DmSourceCreate", params=params, response_reader=self._read_dm_create_source_response)		
		
	def dm_create_application(self, name, source_id, pre_deploy_script=None, post_deploy_script=None):
		"""
		@return ApplicationID[]
		"""
		params = {}
		
		params['Name'] = name
		params['SourceID']  = source_id
		if pre_deploy_script	: params['PreDeployScript'] = pre_deploy_script
		if post_deploy_script	: params['PostDeployScript']= post_deploy_script
		
		return self._request(command="DmApplicationCreate", params=params, response_reader=self._read_dm_create_application_response)	
	
		
	def dm_list_sources(self):
		"""
		@return Source[]
		"""
		return self._request("DmSourcesList", response_reader=self._read_dm_list_sources_response)	
	
	
	def dm_list_applications(self):
		"""
		@return Application[]
		"""
		return self._request("DmApplicationsList", response_reader=self._read_dm_list_applications_response)	
	
	
	def dm_get_deployment_task_log(self, task_id, start_from=None, records_limit=None):
		"""
		@return DMTaskLog[]
		"""
		params = {}
		
		params['DeploymentTaskID'] = task_id
		if start_from	 : params['StartFrom'] 	  = start_from
		if records_limit : params['RecordsLimit'] = records_limit
		
		return self._request(command="DmDeploymentTaskGetLog", params=params, response_reader=self._dm_get_deployment_task_log_response)	
	
	
	def dm_deploy_application(self, app_id, farm_role_id, remote_path):
		"""
		@return DeployTaskResult[]
		"""
		params = {}
		
		params['ApplicationID'] = app_id
		params['FarmRoleID']  = farm_role_id
		params['RemotePath'] = remote_path
		
		return self._request(command="DmApplicationDeploy", params=params, response_reader=self._read_dm_deploy_application_response)		
	
	
	def dm_list_deployment_tasks(self, farm_role_id=None, app_id=None, server_id=None):
		"""
		@return ApplicationID[]
		"""
		params = {}
		
		if farm_role_id	: params['FarmRoleID'] = farm_role_id
		if app_id		: params['ApplicationID'] = app_id
		if server_id	: params['ServerID']= server_id
		
		return self._request(command="DmDeploymentTasksList", params=params, response_reader=self._read_dm_list_deployment_tasks_response)	
	

	def dm_get_deployment_task_status(self, task_id):
		"""
		@return DeployTaskResult[]
		"""
		params = {}
		
		params['DeploymentTaskID'] = task_id
		
		return self._request(command="DmDeploymentTaskGetStatus", params=params, response_reader=self._read_dm_get_deployment_task_status_response)		
		
	def list_apache_virtual_hosts(self):
		"""
		@return VirtualHost[]
		"""
		return self._request("ApacheVhostsList", response_reader=self._read_list_apache_virtual_hosts_response)
	
	
	def list_dns_zones(self):
		"""
		@return DNSZone[]
		"""
		return self._request("DNSZonesList", response_reader=self._read_list_dns_zones_response)

	
	def list_scripts(self):
		"""
		@return Script[]
		"""
		return self._request("ScriptsList", response_reader=self._read_list_scripts_response)
	
	
	def list_farms(self):
		"""
		@return Farm[]
		"""
		return self._request("FarmsList", response_reader=self._read_list_farms_response)
			
			
	def get_farm_status(self, id=None, name=None):
		"""
		@return String
		"""
		assert id or name
		for farm in self.list_farms():
			if farm.id == id or (name and farm.name == name):
				return farm.status
		return None
	
	
	def get_farm_id(self, name):
		"""
		@return String
		"""
		for farm in self.list_farms():
			if farm.name == name:
				return farm.id
		return None
	
	
	def get_farm_name(self, id):
		"""
		@return String
		"""
		for farm in self.list_farms():
			if farm.id == id:
				return farm.name
		return None
	
	
	def get_application_id(self, name):
		"""
		@return String
		"""
		for app in self.dm_list_applications():
			if app.name == name:
				return app.id
		return None
	
	
	def get_application_name(self, name):
		"""
		@return String
		"""
		for app in self.dm_list_applications():
			if app.id == name:
				return app.name
		return None
	
			
	def list_roles(self, platform=None, name=None, prefix=None, image_id=None):
		"""
		@return Role[]
		"""
		params = {}
		
		if platform : params['Platform']= platform
		if name		: params['Name'] 	= name
		if prefix	: params['Prefix'] 	= prefix
		if image_id : params['ImageID'] = image_id
		
		return self._request(command="RolesList", params=params, response_reader=self._read_list_roles_response)


	def list_dns_zone_records(self, zone_name):
		"""
		@return DnsZoneRecord[]
		"""
		params = {'ZoneName':zone_name} if zone_name else {}
		
		return self._request(command="DNSZoneRecordsList", params=params, response_reader=self._read_list_dns_zone_records_response)	
		
	
	def list_events(self, farm_id, start_from=None, records_limit=None):
		"""
		@return Event[]
		"""
		params = {}
		
		params['FarmID'] = farm_id
		if start_from	 : params['StartFrom'] 	  = start_from
		if records_limit : params['RecordsLimit'] = records_limit
		
		return self._request(command="EventsList", params=params, response_reader=self._read_list_events_response)		


	def list_logs(self, farm_id, server_id=None, start_from=None, records_limit=None):
		"""
		@return LogRecord[]
		"""
		params = {}

		params['FarmID'] = farm_id
		if server_id	 : params['ServerID'] = server_id
		if start_from	 : params['StartFrom'] 	  = start_from
		if records_limit : params['RecordsLimit'] = records_limit

		return self._request(command="LogsList", params=params, response_reader=self._read_list_logs_response)


	def scripting_list_logs(self, farm_id, server_id=None, start_from=None, records_limit=None):
		"""
		@return LogRecord[]
		"""
		params = {}

		params['FarmID'] = farm_id
		if server_id	 : params['ServerID'] = server_id
		if start_from	 : params['StartFrom'] 	  = start_from
		if records_limit : params['RecordsLimit'] = records_limit

		return self._request(command="ScriptingLogsList", params=params, response_reader=self._read_scripting_list_logs_response)


	def get_farm_stats(self, farm_id, date=None):
		"""
		@return FarmStat[]
		"""
		params = {}
		
		params['FarmID']= farm_id
		if date: params['Date'] = date
		
		return self._request(command="FarmGetStats", params=params, response_reader=self._read_get_farm_stats_response)	
	
	
	def get_script_details(self,script_id):
		"""
		@return ScriptDetails
		"""
		params = {}
		
		params['ScriptID']= script_id
		
		return self._request(command="ScriptGetDetails", params=params, response_reader=self._read_get_script_details_response)

		
	def get_bundle_task_status(self, bundle_task_id):
		"""
		@return BundleTask
		"""
		params = {}
		
		params['BundleTaskID']= bundle_task_id
		
		return self._request(command="BundleTaskGetStatus", params=params, response_reader=self._read_get_bundle_task_status_response)	


	def get_statistics_graph_URL(self, object_type, object_id, watcher_name, graph_type):
		"""
		@return GraphURL
		"""
		params = {}
		
		params['ObjectType'] = object_type
		params['ObjectID'] = object_id
		params['WatcherName'] = watcher_name
		params['GraphType'] = graph_type
		
		return self._request(command="StatisticsGetGraphURL", params=params, response_reader=self._read_get_statistic_graph_URL_response)	
	

	def terminate_farm(self, farm_id,keep_ebs,keep_eip,keep_dns_zone):
		"""
		@return Result
		"""
		params = {}
		
		params['FarmID'] = farm_id
		params['KeepEBS'] = keep_ebs
		params['KeepEIP'] = keep_eip
		params['KeepDNSZone'] = keep_dns_zone
		
		return self._request(command="FarmTerminate", params=params, response_reader=self._read_terminate_farm_response)	


	def launch_farm(self, farm_id):
		"""
		@return Result
		"""
		params = {}
		
		params['FarmID'] = farm_id
		
		return self._request(command="FarmLaunch", params=params, response_reader=self._read_launch_farm_response)	
	
	
	#not tested
	def create_server_image(self, server_id, role_name):
		"""
		@return BundleTaskID
		"""
		params = {}
		
		params['ServerID']= server_id
		params['RoleName']= role_name
		
		return self._request(command="ServerImageCreate", params=params, response_reader=self._read_create_server_image_response)	
	
		
	def launch_server(self, farm_role_id, increase_max_instances=False):
		"""
		@return ServerID
		"""
		params = {}
		
		params['FarmRoleID']= farm_role_id
		params['IncreaseMaxInstances'] = '1' if increase_max_instances else '0'
		
		return self._request(command="ServerLaunch", params=params, response_reader=self._read_launch_server_response)	



	def create_apache_vhost(self, domain_name,farm_id,farm_role_id,document_root_dir,enable_ssl,ssl_private_key=None,ssl_certificate=None):
		"""
		@return Result
		"""
		params = {}
		
		params['DomainName'] = domain_name
		params['FarmID'] = farm_id
		params['FarmRoleID'] = farm_role_id
		params['DocumentRootDir'] = document_root_dir
		params['EnableSSL'] = enable_ssl
		if ssl_private_key: params['SSLPrivateKey'] = binascii.b2a_base64(ssl_private_key).strip()
		if ssl_certificate: params['SSLCertificate'] = binascii.b2a_base64(ssl_certificate).strip()
		
		return self._request(command="ApacheVhostCreate", params=params, response_reader=self._read_create_apache_vhost_response)			
		
		
	def create_dns_zone(self, domain_name, farm_id=None,farm_role_id=None):
		"""
		@return Result
		"""
		params = {}
		
		params['DomainName']= domain_name
		if farm_id: params['FarmID'] = farm_id
		if farm_role_id: params['FarmRoleID'] = farm_role_id
		
		return self._request(command="DNSZoneCreate", params=params, response_reader=self._read_create_dns_zone_response)			


	def add_dns_zone_record(self, zone_name,type,ttl,record_name,value,priority=None,weight=None,port=None):
		"""
		@return Result
		"""
		params = {}
		
		params['ZoneName'] = zone_name
		params['Type'] = type
		params['TTL'] = ttl
		params['Name'] = record_name
		params['Value'] = value
		if priority: params['Priority'] = priority
		if weight: params['Weight'] = weight
		if port: params['Port'] = port
		
		return self._request(command="DNSZoneRecordAdd", params=params, response_reader=self._read_add_dns_zone_record_response)	


	def remove_dns_zone_record(self, dzone_name, record_id):
		"""
		@return Result
		"""
		params = {}
		
		params['ZoneName']= dzone_name
		params['RecordID'] = record_id
		
		return self._request(command="DNSZoneRecordRemove", params=params, response_reader=self._read_remove_dns_zone_record_response)	


	def reboot_server(self, server_id):
		"""
		@return Result
		"""
		params = {}
		
		params['ServerID'] = server_id
		
		return self._request(command="ServerReboot", params=params, response_reader=self._read_reboot_server_response)		
	

	def terminate_server(self, server_id, decrease_min_instances_setting=None):
		"""
		@return Result
		"""
		params = {}
		
		params['ServerID'] = server_id
		if decrease_min_instances_setting: params['DecreaseMinInstancesSetting'] = decrease_min_instances_setting
		
		return self._request(command="ServerTerminate", params=params, response_reader=self._read_terminate_server_response)	

	
	def get_farm_details(self, farm_id):
		"""
		@return FarmRole[]
		"""
		params = {}
		
		params['FarmID'] = farm_id
		
		return self._request(command="FarmGetDetails", params=params, response_reader=self._read_get_farm_details_response)	
			
	
	def get_farm_role_properties(self, farm_id):
		"""
		@return FarmRoleProperties[]
		"""
		params = {}
		
		params['FarmID'] = farm_id
		
		return self._request(command="FarmGetDetails", params=params, response_reader=self._read_get_farm_role_properties_response)		
	
	
	def list_servers(self, farm_id, farm_role_id=None, columns=None):
		"""
		@return Server[]
		"""
		params = {}
		
		params['FarmID'] = farm_id
		roles = self._request(command="FarmGetDetails", params=params, response_reader=self._read_get_farm_role_properties_response)
		servers = []
		for role in roles:
			if role.servers:
				for server in role.servers:
					server.name = role.name
					server.farm_role_id = role.farm_role_id
					servers.append(server)

		if farm_role_id:
			for server in list(servers):
				if server.farm_role_id != farm_role_id:
					servers.remove(server)

		if columns:
			for server in servers:
				if server.__titles__:
					for variable, column_name in server.__titles__.items():
						if column_name not in columns:
							if column_name in server.__aliases__:
								var = server.__aliases__[column_name]
								if var not in columns:
									del server.__titles__[variable]
									del server.__aliases__[column_name]
							else:
								del server.__titles__[variable]
		return servers


	def remove_farm_role(self, farm_id, farm_role_id):
		"""
		@return Server[]
		"""
		params = {}

		params['FarmID'] = farm_id
		params['FarmRoleID'] = farm_role_id
		return self._request(command="FarmRemoveRole", params=params, response_reader=self._read_remove_farm_role_response)
	
		
	def execute_script(self,farm_id,script_id,timeout,async,farm_role_id=None,server_id=None,config_variables=None,revision=None):
		"""
		@return Result
		"""
		params = {}
		#params['AuthVersion'] = '2'
		params['FarmID'] = farm_id
		params['ScriptID'] = script_id
		params['Timeout'] = timeout
		params['Async'] = async
		if farm_role_id: params['FarmRoleID'] = farm_role_id
		if server_id: params['ServerID'] = server_id
		if config_variables: params['ConfigVariables'] = config_variables
		if revision: params['Revision'] = revision
		
		return self._request(command="ScriptExecute", params=params, response_reader=self._read_execute_script_response)


	def fire_custom_event(self,server_id, event_name, kv_params=None):
		"""
		@return Result
		"""
		params = {}
		params['ServerID'] = server_id
		params['EventName'] = event_name
		if kv_params: params['Params'] = kv_params

		return self._request(command="FireCustomEvent", params=params, response_reader=self._read_fire_custom_event_response)

	
	def list_environments(self):
		"""
		@return Environment[]
		"""
		return self._request("EnvironmentsList", response_reader=self._read_list_environments_response)


	def create_farm(self, name, descr=None, project_id=None, config=None):
		"""
		@return FarmID
		"""
		params = {'Name': name}
		if descr:
			params['Description'] = descr
		if project_id:
			params['ProjectID'] = project_id
		if config:
			params['Configuration'] = config

		return self._request(command="FarmCreate", params=params, response_reader=self._read_create_farm_response)


	def remove_farm(self, farm_id):
		"""
		@return Result
		"""
		params = {'FarmID': farm_id}
		return self._request(command="FarmRemove", params=params, response_reader=self._read_remove_farm_response)


	def add_farm_role(self, farm_id, role_id, platform, location, alias, configuration=None):
		"""
		@return FarmRoleID
		"""
		params = {}
		params['FarmID'] = farm_id
		params['RoleID'] = role_id
		params['Platform'] = platform
		params['CloudLocation'] = location
		if alias:
			params['Alias'] = alias
		if configuration:
			params['Configuration'] = configuration

		return self._request(command="FarmAddRole", params=params, response_reader=self._read_add_farm_fole_response)


	def update_farm_role(self, farm_role_id, alias, configuration=None):
		"""
		@return Result
		"""
		params = {}
		params['FarmRoleID'] = farm_role_id
		if alias:
			params['Alias'] = alias
		if configuration:
			params['Configuration'] = configuration

		return self._request(command="FarmUpdateRole", params=params, response_reader=self._read_update_farm_fole_response)


	def list_global_variables(self, role_id=None, farm_id=None, farm_role_id=None, server_id=None):
		"""
		@return dict
		"""
		params = {}
		if role_id:
			params['RoleID'] = role_id
		if farm_id:
			params['FarmID'] = farm_id
		if farm_role_id:
			params['FarmRoleID'] = farm_role_id
		if server_id:
			params['ServerID'] = server_id

		return self._request("GlobalVariablesList", params=params, response_reader=self._read_list_global_variables_response)


	def get_extended_server_information(self, server_id):
		"""
		@return GlobalVariable[]
		"""
		params = {'ServerID': server_id}

		return self._request("ServerGetExtendedInformation", params=params, response_reader=self._read_get_extended_server_information_response)


	def set_global_variable(self, param_name, param_value, role_id=None, farm_id=None, farm_role_id=None, server_id=None):
		"""
		@return dict
		"""
		params = {"ParamName": param_name, "ParamValue": param_value}
		if role_id:
			params['RoleID'] = role_id
		if farm_id:
			params['FarmID'] = farm_id
		if farm_role_id:
			params['FarmRoleID'] = farm_role_id
		if server_id:
			params['ServerID'] = server_id

		return self._request("GlobalVariableSet", params=params, response_reader=self._read_set_global_variable_response)


	def _read_get_extended_server_information_response(self, xml):
		def from_xml(xml, parent_name=""):
			d = {}
			for element in xml.childNodes:
				if element.nodeType == 3:
					d[parent_name + '/' + element.parentNode.nodeName] = element.nodeValue
				else:
					d.update(from_xml(element, element.parentNode.nodeName))
			return d
		return from_xml(xml)


	def _read_set_global_variable_response(self, xml):
		return self._read_response(xml, node_name='GlobalVariableSetResponse', cls=types.Result, simple_response=True)


	def _read_list_global_variables_response(self, xml):
		return self._read_response(xml, node_name='VariableSet', cls=types.GlobalVariable)


	def _read_update_farm_fole_response(self, xml):
		return self._read_response(xml, node_name='FarmUpdateRoleResponse', cls=types.Result, simple_response=True)


	def _read_add_farm_fole_response(self, xml):
		return self._read_response(xml, node_name='FarmAddRoleResponse', cls=types.FarmRoleID, simple_response=True)


	def _read_remove_farm_role_response(self, xml):
		return self._read_response(xml, node_name='FarmRemoveRoleResponse', cls=types.Result, simple_response=True)


	def _read_remove_farm_response(self, xml):
		return self._read_response(xml, node_name='FarmRemoveResponse', cls=types.Result, simple_response=True)


	def _read_create_farm_response(self, xml):
		return self._read_response(xml, node_name='FarmID', cls=types.FarmID, simple_response=True)


	def _read_list_farm_role_parameters(self, xml):
		return self._read_response(xml, node_name='ParamSet', cls=types.FarmRoleParameter)


	def _read_update_farm_role_parameter(self, xml):
		return self._read_response(xml, node_name='FarmRoleUpdateParameterValueResponse', cls=types.Result, simple_response=True)


	def _read_dm_create_source_response(self, xml):
			return self._read_response(xml, node_name='SourceID', cls=types.SourceID, simple_response=True)

	
	def _read_dm_create_application_response(self, xml):
		return self._read_response(xml, node_name='ApplicationID', cls=types.ApplicationID, simple_response=True)
	
	
	def _read_dm_deploy_application_response(self, xml):
		return self._read_response(xml, node_name='DeploymentTasksSet', cls=types.DeploymentTaskResult)

	
	def _read_dm_list_deployment_tasks_response(self, xml):
		return self._read_response(xml, node_name='DeploymentTasksSet', cls=types.DeploymentTaskResult)		
	
	
	def _read_dm_list_sources_response(self, xml):
		return self._read_response(xml, node_name='SourceSet', cls=types.Source)			

	
	def _dm_get_deployment_task_log_response(self, xml):
		return self._read_response(xml, node_name='LogSet', cls=types.DMTaskLog, wrap_page=True)


	def _read_dm_get_deployment_task_status_response(self, xml):
		return self._read_response(xml, node_name='DeploymentTaskStatus', cls=types.DeploymentTask, simple_response=True)	
	

	def _read_dm_list_applications_response(self, xml):
		return self._read_response(xml, node_name='ApplicationSet', cls=types.Application)
	
	
	def _read_get_script_details_response(self, xml):
		return self._read_response(xml, node_name='ScriptRevisionSet', cls=types.ScriptRevision)
	
			
	def _read_list_apache_virtual_hosts_response(self, xml):
		return self._read_response(xml, node_name='ApacheVhostSet', cls=types.VirtualHost)
		
	
	def _read_list_dns_zones_response(self, xml):
		return self._read_response(xml, node_name='DNSZoneSet', cls=types.DNSZone)
	
	
	def _read_list_scripts_response(self, xml):
		return self._read_response(xml, node_name='ScriptSet', cls=types.Script)
	
	
	def _read_list_farms_response(self, xml):
		return self._read_response(xml, node_name='FarmSet', cls=types.Farm)
	
	
	def _read_list_roles_response(self, xml):
		return self._read_response(xml, node_name='RoleSet', cls=types.Role)
	
	
	def _read_list_dns_zone_records_response(self, xml):
		return self._read_response(xml, node_name='ZoneRecordSet', cls=types.DnsZoneRecord)
	
	
	def _read_list_events_response(self, xml):
		return self._read_response(xml, node_name='EventSet', cls=types.Event, wrap_page=True)


	def _read_list_logs_response(self, xml):
		return self._read_response(xml, node_name='LogSet', cls=types.LogRecord, wrap_page=True)


	def _read_scripting_list_logs_response(self, xml):
		return self._read_response(xml, node_name='LogSet', cls=types.ScriptingLogRecord, wrap_page=True)


	def _read_get_farm_stats_response(self, xml):
		return self._read_response(xml, node_name='StatisticsSet', cls=types.FarmStat)
	

	def _read_get_farm_details_response(self, xml):
		return self._read_response(xml, node_name='FarmRoleSet', cls=types.FarmRole)
	
	
	def _read_get_farm_role_properties_response(self, xml):
		return self._read_response(xml, node_name='FarmRoleSet', cls=types.FarmRoleProperties)
	
	
	def _read_list_servers_response(self, xml):
		return self._read_response(xml, node_name='ServerSet', cls=types.Server) # FarmRoleSet
	
		
	def _read_get_bundle_task_status_response(self, xml):
		return self._read_response(xml, node_name='BundleTaskGetStatusResponse', cls=types.BundleTask, simple_response=True)
	
	
	def _read_create_server_image_response(self, xml):
		return self._read_response(xml, node_name='ServerCreateImageResponse', cls=types.BundleTaskID, simple_response=True)
	
	
	def _read_launch_server_response(self, xml):
		return self._read_response(xml, node_name='ServerLaunchResponse', cls=types.ServerID, simple_response=True)
	
	
	def _read_get_statistic_graph_URL_response(self, xml):
		return self._read_response(xml, node_name='StatisticsGetGraphURLResponse', cls=types.GraphURL, simple_response=True)
	
	
	def _read_create_apache_vhost_response(self, xml):
		return self._read_response(xml, node_name='ApacheVhostCreate', cls=types.Result, simple_response=True)
	
	
	def _read_create_dns_zone_response(self, xml):
		return self._read_response(xml, node_name='DNSZoneCreateResponse', cls=types.Result, simple_response=True)
	
	
	def _read_add_dns_zone_record_response(self, xml):
		return self._read_response(xml, node_name='DNSZoneRecordAddResponse', cls=types.Result, simple_response=True)
	

	def _read_remove_dns_zone_record_response(self, xml):
		return self._read_response(xml, node_name='DNSZoneRecodRemoveResponse', cls=types.Result, simple_response=True)
	

	def _read_terminate_farm_response(self, xml):
		return self._read_response(xml, node_name='FarmTerminateResponse', cls=types.Result, simple_response=True)
	
	
	def _read_launch_farm_response(self, xml):
		return self._read_response(xml, node_name='FarmLaunchResponse', cls=types.Result, simple_response=True)


	def _read_reboot_server_response(self, xml):
		return self._read_response(xml, node_name='ServerRebootResponse', cls=types.Result, simple_response=True)	
	
	
	def _read_terminate_server_response(self, xml):
		return self._read_response(xml, node_name='TerminateInstanceResponse', cls=types.Result, simple_response=True)		
	
	
	def _read_execute_script_response(self, xml):
		return self._read_response(xml, node_name='ExecuteScriptResponse', cls=types.Result, simple_response=True)	


	def _read_fire_custom_event_response(self, xml):
		return self._read_response(xml, node_name='FireCustomEventResponse', cls=types.EventID, simple_response=True)
	
	
	def _read_list_environments_response(self, xml):
		return self._read_response(xml, node_name='EnvironmentSet', cls=types.Environment)
		

	def _read_response(self, xml, node_name, cls, wrap_page=False, simple_response=False):
		response = xml.documentElement
		ret = list()
		
		if not self.is_document_empty(response):
			self.last_transaction_id = self.get_transaction_id(response)
			
			if response.nodeName == 'Error':
				raise ScalrAPIError.fromxml(response)

			if simple_response:
				ret.append(cls.fromxml(response))
			else:
				#utterly experimental, can ruin everything
				for node in xml.getElementsByTagName(node_name):
					for child in node.childNodes:
						ret.append(cls.fromxml(child))

			if wrap_page:
				page = types.Page.fromxml(response)
				page.scalr_objects = ret
				return page
		#print ret
		return ret

				
	def _request (self, command, params={}, response_reader=None, response_reader_args=None):
		xml = self.fetch(command, **params)
		response_reader_args = response_reader_args or ()
		return response_reader(xml, *response_reader_args)
	
		
	def get_transaction_id(self, response):
		id = None
		for child in response.childNodes:
			if 'TransactionID' == child.nodeName:
				return child.firstChild.nodeValue.strip()
		return id
	
	
	def is_document_empty(self, response):
		return not response.firstChild.hasChildNodes() if response and response.firstChild else False

	
class ScalrAPIError(Exception):
	
	@classmethod
	def fromxml(cls, xml):
		child = types.get_items_first_child(xml, 'Message')
		msg = child.nodeValue.strip() if child else 'Cannot parse document'
		return ScalrAPIError(msg)
	
		
def xml_strip(el):
	for child in list(el.childNodes):
		if child.nodeType==child.TEXT_NODE and child.nodeValue.strip() == '':
			el.removeChild(child)
		else:
			xml_strip(child)
	return el	


def sign_http_request_v3(data, key_id, access_key, timestamp=None):
	date = time.strftime("%Y-%m-%dT%H:%M:%S.000Z", timestamp or time.gmtime())
	data['TimeStamp']=date
	canonical_string = 	"%s:%s:%s" % (data['Action'], key_id, data['TimeStamp'])
	digest = hmac.new(access_key, canonical_string, hashlib.sha256).digest()
	sign = binascii.b2a_base64(digest).strip()
	return sign, date		

		
def sign_http_request_v1(data, key, timestamp=None):
	date = time.strftime("%Y-%m-%dT%H:%M:%S.000Z", timestamp or time.gmtime())
	data['TimeStamp']=date
	canonical_string = get_canonical_string(data) if hasattr(data, "__iter__") else data
	digest = hmac.new(key, canonical_string, hashlib.sha256).digest()
	sign = binascii.b2a_base64(digest).strip()
	
	return sign, date


def get_canonical_string (params={}):
	s = ""
	for key in sorted(params.keys(), key=str.lower):
		s = s + str(key) + str(params[key])
	return s



from operator import itemgetter as _itemgetter
from keyword import iskeyword as _iskeyword
from _abcoll import MutableMapping

def namedtuple(typename, field_names, verbose=False, rename=False):
    """Returns a new subclass of tuple with named fields.

    >>> Point = namedtuple('Point', 'x y')
    >>> Point.__doc__                   # docstring for the new class
    'Point(x, y)'
    >>> p = Point(11, y=22)             # instantiate with positional args or keywords
    >>> p[0] + p[1]                     # indexable like a plain tuple
    33
    >>> x, y = p                        # unpack like a regular tuple
    >>> x, y
    (11, 22)
    >>> p.x + p.y                       # fields also accessable by name
    33
    >>> d = p._asdict()                 # convert to a dictionary
    >>> d['x']
    11
    >>> Point(**d)                      # convert from a dictionary
    Point(x=11, y=22)
    >>> p._replace(x=100)               # _replace() is like str.replace() but targets named fields
    Point(x=100, y=22)

    """

    # Parse and validate the field names.  Validation serves two purposes,
    # generating informative error messages and preventing template injection attacks.
    if isinstance(field_names, basestring):
        field_names = field_names.replace(',', ' ').split() # names separated by whitespace and/or commas
    field_names = tuple(map(str, field_names))
    if rename:
        names = list(field_names)
        seen = set()
        for i, name in enumerate(names):
            if (not min(c.isalnum() or c=='_' for c in name) or _iskeyword(name)
                or not name or name[0].isdigit() or name.startswith('_')
                or name in seen):
                    names[i] = '_%d' % i
            seen.add(name)
        field_names = tuple(names)
    for name in (typename,) + field_names:
        if not min(c.isalnum() or c=='_' for c in name):
            raise ValueError('Type names and field names can only contain alphanumeric characters and underscores: %r' % name)
        if _iskeyword(name):
            raise ValueError('Type names and field names cannot be a keyword: %r' % name)
        if name[0].isdigit():
            raise ValueError('Type names and field names cannot start with a number: %r' % name)
    seen_names = set()
    for name in field_names:
        if name.startswith('_') and not rename:
            raise ValueError('Field names cannot start with an underscore: %r' % name)
        if name in seen_names:
            raise ValueError('Encountered duplicate field name: %r' % name)
        seen_names.add(name)

    # Create and fill-in the class template
    numfields = len(field_names)
    argtxt = repr(field_names).replace("'", "")[1:-1]   # tuple repr without parens or quotes
    reprtxt = ', '.join('%s=%%r' % name for name in field_names)
    template = '''class %(typename)s(tuple):
        '%(typename)s(%(argtxt)s)' \n
        __slots__ = () \n
        _fields = %(field_names)r \n
        def __new__(_cls, %(argtxt)s):
            return _tuple.__new__(_cls, (%(argtxt)s)) \n
        @classmethod
        def _make(cls, iterable, new=tuple.__new__, len=len):
            'Make a new %(typename)s object from a sequence or iterable'
            result = new(cls, iterable)
            if len(result) != %(numfields)d:
                raise TypeError('Expected %(numfields)d arguments, got %%d' %% len(result))
            return result \n
        def __repr__(self):
            return '%(typename)s(%(reprtxt)s)' %% self \n
        def _asdict(self):
            'Return a new dict which maps field names to their values'
            return dict(zip(self._fields, self)) \n
        def _replace(_self, **kwds):
            'Return a new %(typename)s object replacing specified fields with new values'
            result = _self._make(map(kwds.pop, %(field_names)r, _self))
            if kwds:
                raise ValueError('Got unexpected field names: %%r' %% kwds.keys())
            return result \n
        def __getnewargs__(self):
            return tuple(self) \n\n''' % locals()
    for i, name in enumerate(field_names):
        template += '        %s = _property(_itemgetter(%d))\n' % (name, i)
    if verbose:
        print template

    # Execute the template string in a temporary namespace
    namespace = dict(_itemgetter=_itemgetter, __name__='namedtuple_%s' % typename,
                     _property=property, _tuple=tuple)
    try:
        exec template in namespace
    except SyntaxError, e:
        raise SyntaxError(e.message + ':\n' + template)
    result = namespace[typename]

    # For pickling to work, the __module__ variable needs to be set to the frame
    # where the named tuple is created.  Bypass this step in enviroments where
    # sys._getframe is not defined (Jython for example) or sys._getframe is not
    # defined for arguments greater than 0 (IronPython).
    
    #try:
    #    result.__module__ = _sys._getframe(1).f_globals.get('__name__', '__main__')
    #except (AttributeError, ValueError):
    #    pass

    return result
   
   
class OrderedDict(dict, MutableMapping):
    'Dictionary that remembers insertion order'
    # An inherited dict maps keys to values.
    # The inherited dict provides __getitem__, __len__, __contains__, and get.
    # The remaining methods are order-aware.
    # Big-O running times for all methods are the same as for regular dictionaries.

    # The internal self.__map dictionary maps keys to links in a doubly linked list.
    # The circular doubly linked list starts and ends with a sentinel element.
    # The sentinel element never gets deleted (this simplifies the algorithm).
    # Each link is stored as a list of length three:  [PREV, NEXT, KEY].

    def __init__(self, *args, **kwds):
        '''Initialize an ordered dictionary.  Signature is the same as for
        regular dictionaries, but keyword arguments are not recommended
        because their insertion order is arbitrary.

        '''
        if len(args) > 1:
            raise TypeError('expected at most 1 arguments, got %d' % len(args))
        try:
            self.__root
        except AttributeError:
            self.__root = root = [None, None, None]     # sentinel node
            PREV = 0
            NEXT = 1
            root[PREV] = root[NEXT] = root
            self.__map = {}
        self.update(*args, **kwds)

    def __setitem__(self, key, value, PREV=0, NEXT=1, dict_setitem=dict.__setitem__):
        'od.__setitem__(i, y) <==> od[i]=y'
        # Setting a new item creates a new link which goes at the end of the linked
        # list, and the inherited dictionary is updated with the new key/value pair.
        if key not in self:
            root = self.__root
            last = root[PREV]
            last[NEXT] = root[PREV] = self.__map[key] = [last, root, key]
        dict_setitem(self, key, value)

    def __delitem__(self, key, PREV=0, NEXT=1, dict_delitem=dict.__delitem__):
        'od.__delitem__(y) <==> del od[y]'
        # Deleting an existing item uses self.__map to find the link which is
        # then removed by updating the links in the predecessor and successor nodes.
        dict_delitem(self, key)
        link = self.__map.pop(key)
        link_prev = link[PREV]
        link_next = link[NEXT]
        link_prev[NEXT] = link_next
        link_next[PREV] = link_prev

    def __iter__(self, NEXT=1, KEY=2):
        'od.__iter__() <==> iter(od)'
        # Traverse the linked list in order.
        root = self.__root
        curr = root[NEXT]
        while curr is not root:
            yield curr[KEY]
            curr = curr[NEXT]

    def __reversed__(self, PREV=0, KEY=2):
        'od.__reversed__() <==> reversed(od)'
        # Traverse the linked list in reverse order.
        root = self.__root
        curr = root[PREV]
        while curr is not root:
            yield curr[KEY]
            curr = curr[PREV]

    def __reduce__(self):
        'Return state information for pickling'
        items = [[k, self[k]] for k in self]
        tmp = self.__map, self.__root
        del self.__map, self.__root
        inst_dict = vars(self).copy()
        self.__map, self.__root = tmp
        if inst_dict:
            return (self.__class__, (items,), inst_dict)
        return self.__class__, (items,)

    def clear(self):
        'od.clear() -> None.  Remove all items from od.'
        try:
            for node in self.__map.itervalues():
                del node[:]
            self.__root[:] = [self.__root, self.__root, None]
            self.__map.clear()
        except AttributeError:
            pass
        dict.clear(self)

    setdefault = MutableMapping.setdefault
    update = MutableMapping.update
    pop = MutableMapping.pop
    keys = MutableMapping.keys
    values = MutableMapping.values
    items = MutableMapping.items
    iterkeys = MutableMapping.iterkeys
    itervalues = MutableMapping.itervalues
    iteritems = MutableMapping.iteritems
    __ne__ = MutableMapping.__ne__

    def popitem(self, last=True):
        '''od.popitem() -> (k, v), return and remove a (key, value) pair.
        Pairs are returned in LIFO order if last is true or FIFO order if false.

        '''
        if not self:
            raise KeyError('dictionary is empty')
        key = next(reversed(self) if last else iter(self))
        value = self.pop(key)
        return key, value

    def __repr__(self):
        'od.__repr__() <==> repr(od)'
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, self.items())

    def copy(self):
        'od.copy() -> a shallow copy of od'
        return self.__class__(self)

    @classmethod
    def fromkeys(cls, iterable, value=None):
        '''OD.fromkeys(S[, v]) -> New ordered dictionary with keys from S
        and values equal to v (which defaults to None).

        '''
        d = cls()
        for key in iterable:
            d[key] = value
        return d

    def __eq__(self, other):
        '''od.__eq__(y) <==> od==y.  Comparison to another OD is order-sensitive
        while comparison to a regular mapping is order-insensitive.

        '''
        if isinstance(other, OrderedDict):
            return len(self)==len(other) and \
                   all(_imap(_eq, self.iteritems(), other.iteritems()))
        return dict.__eq__(self, other)

    def __del__(self):
        self.clear()                # eliminate cyclical references



'''
Created on Feb 21th, 2011

@author: Dmytro Korsakov
'''
import os

from ConfigParser import ConfigParser, NoSectionError, NoOptionError

from prettytable import PrettyTable	


class ScalrCfgError(BaseException):
	pass


class ScalrEnvError(ScalrCfgError):
	pass

	
class ConfigSection(object):
	config_name = None
	options = {}
	
	def __init__(self, **kwargs):
		for arg in kwargs:
			if hasattr(self, arg):
				setattr(self, arg, kwargs[arg])

	def write(self, base_path, section):
		config = ConfigParser()
		
		path = os.path.join(base_path, self.config_name)
		if os.path.exists(path):
			config.read(path)
		elif not os.path.exists(os.path.dirname(path)):
			os.makedirs(os.path.dirname(path))
			
		if not config.has_section(section):
			config.add_section(section)
		
		for option in self.options:
			config.set(section, self.options[option], getattr(self, option))
			
		file = open(path, 'w')
		config.write(file)
		file.close()
			
	@classmethod
	def from_ini(cls, base_path, section):
		obj = cls()
		path = os.path.join(base_path, cls.config_name)
		
		if os.path.exists(path):
			config = ConfigParser()	
			config.read(path)
			setattr(obj, 'name', section)
		
			for option in cls.options:
				try:
					value = config.get(section, cls.options[option])
					setattr(obj, option, value)
				except (NoSectionError, NoOptionError), e:
						continue
		return obj

	
class Environment(ConfigSection):
	url=None
	key_id=None
	key=None
	env_id = None
	api_version = "2.3.0"
	auth_type = None
	ldap_login = None
	ldap_password = None
	
	config_name = 'config.ini'
	
	options = dict(
			url = 'scalr_url',
			key_id = 'scalr_key_id',
			key = 'scalr_api_key',
			env_id = 'env_id',
			api_version = 'version',
			auth_type = 'auth_type',
			ldap_login = 'ldap_login',
			#ldap_password = 'ldap_password'
	)
	
	def write(self, base_path, section='api'):
		super(Environment, self).write(base_path, section)
			
	@classmethod
	def from_ini(cls, base_path, section='api'):
		return super(Environment, cls).from_ini(base_path, section)
	
	def __repr__(self):
		column_names = ('setting','value')
		table = PrettyTable(column_names)
		#prettytable 0.5/0.6 support
		if hasattr(table, 'align'):
			table.align = 'l'
		else:
			for field in column_names:
				table.set_field_align(field, 'l')		
		
		visible_length = 26
		
		table.add_row(('url', self.url))
		table.add_row(('access key', self.key[:visible_length]+'...' if len(self.key)>40 else self.key))
		table.add_row(('key id', self.key_id))
		table.add_row(('environment id', self.env_id))
		table.add_row(('version', self.api_version))
		table.add_row(('ldap_login', self.ldap_login))
		#table.add_row(('ldap_password', self.ldap_password))
		table.add_row(('auth_type', self.auth_type))

		
		res = str(table)
		return res
	

class Configuration:
	logger = None
	base_path = None
	
	environment = None
	application = None
	repository = None
	scripts = None
	
	def __init__(self, base_path=None, logger=None):
		self.base_path = base_path or os.path.expanduser("~/.scalr/")
		self.logger = logger

	def set_logger(self, logger):
		self.logger = logger

	def set_environment(self, key=None, key_id=None, url=None, env_id=None, ldap_login=None, ldap_password=None):
		self.environment = Environment.from_ini(self.base_path)
		if url:
			self.environment.url = url
		if env_id:
			self.environment.env_id = env_id

		if ldap_login:
			self.environment.ldap_login = ldap_login
			if not self.environment.auth_type:
				#XXX: Ugly
				self.environment.auth_type = 'ldap'

		if ldap_password:
			self.environment.ldap_password = ldap_password

		if key:
			self.environment.key = key
			if not self.environment.auth_type:
				#XXX: Ugly
				self.environment.auth_type = 'password'
		if key_id:
			self.environment.key_id = key_id


		if not self.environment.key or not self.environment.key_id or not self.environment.url:
			raise ScalrEnvError('Environment not set.')




'''
Created on Feb 21th, 2011

@author: Dmytro Korsakov
'''

from prettytable import PrettyTable
from types_ import DynamicScalrObject, Page
class TableViewer:
	
	data = None
	
	def __str__(self):
		lines = []
		for table,text in self.data.items():
			line = ''
			
			if text:
				line += '%s\n' % text
				
			if isinstance(table, PrettyTable):
				line += '%s\n' % table
				
			lines.append(line)
		return '\n'.join(lines)
	
	
	def __init__(self, response):
		
		'''
		data = @dict(table=title)
		'''
		self.data = {}
		
		if response:
				
			if isinstance(response, list):
				self.data[self.prepare_table(response)] = None
				
			#for dynamic key-value objects
			elif isinstance(response, DynamicScalrObject):
				pt = PrettyTable([response.__title__, 'Value'])
				for k,v in response.data:
					pt.add_row([k,v])
				self.data[pt] = None
			
			#for paginated objects	
			elif isinstance(response, Page):
				objects = response.scalr_objects
				plain_text = 'Total records: %s\nStart:%s\nLimit:%s\n' % (
						response.total_records, 
						response.start_from, 
						response.records_limit)
				self.data[self.prepare_table(objects) if objects else response.total_records] = plain_text
				

	def apply_aliases(self, column_names, aliases):
		if aliases: 
			
			for key in aliases:
				if key in column_names:
					column_names[column_names.index(key)] = aliases[key]
				
		return column_names
	
	
	def get_column_names(self, objects):
		column_names = []

		if objects:
			object = objects[0]
			column_names = object.__titles__.values()
			aliases = object.__aliases__
		
			return self.apply_aliases(column_names, aliases)
			
		
	def prepare_table(self, objects):		
		column_names = self.get_column_names(objects)
		
		pt = PrettyTable(column_names, caching=False)
		
		#set left allign to all columns
		#prettytable 0.5/0.6 support
		if hasattr(pt, 'align'):
			pt.align = 'l'
		else:
			for field in column_names:
				pt.set_field_align(field, 'l')
		
		#filling the table 
		for scalr_obj in objects:
			row = []
			for attribute in scalr_obj.__titles__.keys():
				cell = getattr(scalr_obj, attribute)
				
				#packing list into cell 
				if isinstance(cell, list):
					cell = ';'.join(cell)
					
				#packing dict into cell	
				elif isinstance(cell, dict):
					cell = ';'.join(['%s=%s'%(k,v) for k,v in cell.items()])
					
				row.append(cell)
			pt.add_row(row)
			
		return pt


'''
Created on Feb 21th, 2011

@author: Dmytro Korsakov
'''
import os
from setuptools import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

description = "Scalr-tools is a command-line interface to your Scalr account"

cfg = dict(
	name = "scalr",
	version = read("VERSION").strip(),
	description = description,
	author = "Scalr Inc.",
	author_email = "dmitry@scalr.com",
	url = "https://scalr.net",
	license = "GPL",
	platforms = "any",
	packages = find_packages(),
	include_package_data = True,
	scripts=['bin/scalr'],
	install_requires = ["prettytable"],
	long_description=read('README'),
	classifiers=[
		'Programming Language :: Python',
		'Programming Language :: Python :: 2',
		'License :: OSI Approved :: GNU General Public License (GPL)',
		'Operating System :: OS Independent',
		'Development Status :: 4 - Beta',
		'Environment :: Console',
		'Intended Audience :: System Administrators',
		'Topic :: Utilities'
		]
	)
setup(**cfg)
